{% extends "base.html" %}

{% block body %}
<h3>class based view via render_table</h3>
<pre>{{ "{%" }} load django_tables2 {{ "%}" }}
{{ "{%" }} render_table example3 {{ "%}" }}</pre>
{% load django_tables2 %}
{% render_table table %}
{% endblock %}
